import java.util.Collections;
import java.util.ArrayList;
import java.util.List;

/**
 * Contains the challenge problems and code to solve.
 * 
 * @author Ed Parrish
 * @version 1.0  5/4/2014
 */
public class Challenge {
    private String problem;
    private int probValue;
    private List<CodeLine> code = new ArrayList<CodeLine>();
    private Question question = new Question();

    /**
     * Constructor for objects of class Challenges
     */
    public Challenge(String problemStatement) {
        this (problemStatement, 0);
    }

    public Challenge(String problemStatement, int problemValue) {
        problem = problemStatement;
        probValue = problemValue;
    }

    /**
     * Add a CodeLine to this challenge.
     * 
     * @param line the CodeLine to add.
     */
    public void addCodeLine(CodeLine line) {
        code.add(line);
    }

    /**
     * Returns the problem text for this Challenge.
     *
     * @return The problem text.
     */
    public String getProblem() {
        return problem;
    }

    /**
     * Returns the problem point value for this Challenge.
     *
     * @return The problem point value.
     */
    public int getProblemValue() {
        return probValue;
    }
    
    /**
     * Returns the sum of the problem point value and question value for this 
     * Challenge.
     *
     * @return The sum of the problem point value and question value.
     */
    public int getTotalValue() {
        return probValue + question.getValue();
    }
    
    /**
     * Return the lines of code for this Challenge.
     *
     * @return The lines of code.
     */
    public List<CodeLine> getCodeLines() {
        return new ArrayList(code);
    }

    /**
     * Set the text of the optional question for this Challenge.
     *
     * @param text The text of the question.
     */
    public void setQuestionText(String text) {
        question.setText(text);
    }

    /**
     * Tests if this Challenge has an optional question.
     *
     * @return <code>true</code> if this Challenge has a question;
     *         otherwise <code>false</code>.
     */
    public boolean hasQuestion() {
        return question != null && question.hasText();
    }
    
    /**
     * Set the value of the optional question for this Challenge.
     *
     * @param value The point value of the question.
     */
    public void setQuestionValue(int value) {
        question.setValue(value);
    }

    /**
     * Returns a copy of the optional Question if it exists.
     *
     * @return a copy of the optional Question if it exists; otherwise returns null. 
     */
    public Question getQuestion() {
        Question q = null;
        if (hasQuestion()) q = new Question(question);
        return q;
    }

    /**
     * Adds an answer for the question of this Challenge.
     *
     * @param answerText The answer text.
     * @param correct Set <code>true</code> if the answer is correct;
     *        otherwise <code>false</code>.
     */
    public void addAnswer(String answerText, boolean correct) {
        question.addAnswer(answerText, correct);
    }

    /**
     * Randomly permutes the list of answers.
     */
    public void shuffleAnswers() {
        question.shuffleAnswers();
    }

    /**
     * Set the optional explanation for this question.
     *
     * @param text The explanation for this question.
     */
    public void setExplanation(String text) {
        question.setExplanation(text);
    }

    /**
     * Returns this challenge as a string.
     *
     * @return The string value of this challenge.
     */
    public String toString() {
        String str = problem + "(" + probValue + ")\n";
        for (CodeLine line : code) {
            str += line.toString() + "\n";
        }
        if (question.hasText()) {
            str += question.toString();
        }
        return str;
    }
}
