import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Font;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * A line of text.
 * 
 * @author Ed Parrish 
 * @version 1.0  4/27/2014
 */
public class CodeLine extends Actor implements Comparable<CodeLine>
{
    private static final int LEFT = 0;
    private static final int UP = 1;
    private static final int RIGHT = 2;
    private static final int DOWN = 3;
    private final int WIDTH = (Game.WIDTH - 1) / 2 * 20 - 10;
    public static final Font LINE_FONT = new Font("SansSerif", Font.PLAIN, 14);
    //private static final Color CODE_BG = new Color(250, 240, 230, 192);
    //private static final Color CODE_BG = new Color(160, 228, 247, 128);
    private static final Color CODE_BG = new Color(240, 240, 255, 160);
    private static final Color ERROR_BG = new Color(255, 0, 0, 48);
    private static final Color VALID_BG = new Color(0, 255, 0, 48);
    // Track text, indentation and acceptable index range
    private String text;
    private int indentLevel;
    private int lineStart;
    private int lineEnd;
    private static GreenfootImage[] arrows;
    private static GreenfootImage delete = new GreenfootImage("delete.png");
    private Game gm;

    public CodeLine(String text, int indentation, int startLine, int endLine) {
        this.text = text;
        indentLevel = indentation;
        lineStart = startLine;
        lineEnd = endLine;
        initialize();
        makeImage(CODE_BG);
    }

    public static void initialize() {
        if (arrows == null)
        {
            arrows = new GreenfootImage[4];
            arrows[0] = new GreenfootImage("arrow16x16.png");
            for (int i = 1; i < arrows.length; i++)
            {
                arrows[i] = new GreenfootImage(arrows[0]);
                arrows[i].rotate(90 * i);
            }
        }
    }

    protected void addedToWorld(World world) {
        gm = (Game) getWorld();
    }

    /**
     * Handle drag and drop code.
     */
    public void act() {
        // Drag and drop code
        if (gm.canDrag() && Greenfoot.mouseDragged(this)) {
            MouseInfo mouse = Greenfoot.getMouseInfo();
            int x = mouse.getX();
            int y = mouse.getY();
            if (x < gm.LEFT_X) x = gm.LEFT_X;
            if (y < gm.CODE_Y) y = gm.CODE_Y;
            if (y > gm.getBottom()) y = gm.getBottom();
            setLocation(x, y);             
        }
        if (Greenfoot.mouseDragEnded(this) && getX() < gm.RIGHT_X - WIDTH / 40) {
            Greenfoot.playSound("drop.wav");
            // Make sure not on top of another CodeLine
            List<CodeLine> lines = gm.getObjectsAt(getX(), getY(), CodeLine.class);
            if (lines.size() > 1) {
                insert();
            }
        }
    }    

    /**
     * Insert this CodeLine if overlapping with another CodeLine.
     */
    private void insert() {
        // Restrict code to only those lines in the ordered area
        List<CodeLine> allLines = gm.getObjects(CodeLine.class);
        ArrayList<CodeLine> list = new ArrayList<CodeLine>();
        for (int i = 0; i < allLines.size(); i++) {
            CodeLine line = allLines.get(i);
            if (line.getX() < gm.RIGHT_X - 1 && line != this) {
                list.add(line);
            }
        }
        Collections.sort(list);
        // Move up if we can
        int bestY = gm.CODE_Y;
        boolean hasMoved = false;
        for (int i = 0; i < list.size(); i++) {
            CodeLine line = list.get(i);
            if (line.getY() > bestY && line.getY() <= getY()) {
                line.setLocation(line.getX(), line.getY() - 1);
                hasMoved = true;
            }
            bestY++;
        }
        if (hasMoved) return;
        // Otherwise move down from the overlap y
        int lastY = getY() - 1;
        for (int i = getY() - gm.CODE_Y; i < list.size(); i++) {
            CodeLine line = list.get(i);
            if (line.getY() != lastY + 1) {
                return; // prevent bunching
            }
            line.setLocation(line.getX(), line.getY() + 1);
            lastY++;
        }
    }

    /**
     * Compares this object with the specified object for order.
     *
     * @param tl the other object
     * @return a negative integer, zero, or a positive integer as this object is
     * less than, equal to, or greater than the specified object.
     */
    public int compareTo(CodeLine cl) {
        return getY() - cl.getY();
    }

    /**
     * Returns the text of this CodeLine.
     *
     * @return The text value.
     */
    public String getText() {
        return text;
    }
    
    public int getIndentationLevel() {
        return indentLevel;
    }
    
    public boolean isNeeded() {
       return lineStart > 0 && lineEnd > 0;
    } 

    public boolean checkLine(int line) {
        // Check for unneeded code
        if (getX() < ((Game)getWorld()).RIGHT_X && !isNeeded()) {
            makeImage(ERROR_BG);
            GreenfootImage img = getImage();
            img.drawImage(delete, img.getWidth() - 16, 0);
            return false;
        }
        // Exit if this line is unneeded code still in the pile
        if (!isNeeded()) return true;
        // Check if in wrong pile
        if (getX() >= ((Game)getWorld()).RIGHT_X) {
            makeImage(ERROR_BG);
            addArrow(LEFT);
            return false;
        }
        // Check for wrong line order
        if (line < lineStart || line > lineEnd) {
            //             System.out.println("Error: wrong line " + line + " s/b " 
            //                 + lineStart + "-" + lineEnd);
            makeImage(ERROR_BG);
            if (line < lineStart) {
                addArrow(DOWN);
            } else {
                addArrow(UP);
            }
            return false;
        }
        // Check for indentation error
        if (getX() - ((Game)getWorld()).LEFT_X != indentLevel) {
            //             System.out.println("Error: wrong indentation " + line);
            makeImage(ERROR_BG);
            if (getX() - ((Game)getWorld()).LEFT_X > indentLevel) {
                addArrow(LEFT);
            } else {
                addArrow(RIGHT);
            }
            return false;
        }
        makeImage(VALID_BG);
        return true;
    }

    private void addArrow(int num) {
        GreenfootImage img = getImage();
        img.drawImage(arrows[num], img.getWidth() - 16, 0);
    }

    private void makeImage(Color bg) {
        // All lines of text are same size
        GreenfootImage img = new GreenfootImage(WIDTH, 20);
        img.setColor(bg);
        img.fill();
        img.setColor(Color.BLACK);
        img.setFont(LINE_FONT);
        img.drawString(text, 2, 16);
        setImage(img);
    } 

    public String toString() {
        return "("+text+" y:"+getY()+" indent:"+indentLevel+" start:"+lineStart+" end:"+lineEnd+")";
    }
}
