import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.io.*;
import java.util.Collections;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import java.awt.Graphics; // testing GUIComponent.wordWrap()

/**
 * Manages the operation of the game.
 * 
 * @author Ed Parrish 
 * @version 1.0  5/9/2014
 */
public class Game extends World implements ActionListener
{
    public static final int WIDTH = 35; // must be an odd number for some reason.
    public static final int HEIGHT = 27;
    public static final int CELL_SIZE = 20;
    public static final int CODE_Y = 10;
    public static final int CODE_HEIGHT = 12;
    public final int LEFT_X;
    public final int RIGHT_X;
    //public static final Font LARGE_FONT = new Font("SansSerif", Font.BOLD, 20);
    public static final Font MED_FONT = new Font("SansSerif", Font.BOLD, 16);
    public static final Font PROB_FONT = new Font("SansSerif", Font.BOLD, 16);
    public static final GreenfootImage BTN_UP = new GreenfootImage("roundedbtn-up.png");
    public static final GreenfootImage BTN_DWN = new GreenfootImage("roundedbtn-down.png");
    public static final GreenfootImage BTN_HVR = new GreenfootImage("roundedbtn-hover.png");

    private static final String INSTR_START = "Press the Start button to begin the challenge";
    private static final String INSTR_TEXT = "Order the statements by dragging them "
        + "to the left side.\nIndent statements inside braces.";
    private static final String INSTR_NEXT = "Press Next to go on to the next challenge.";
    private static final String INSTR_NEXT_Q = "Good! Now answer the bonus question."
        +"\nThen press Next to go on to the next challenge.";
    private static final String INSTR_ERROR = "Correct the errors by moving lines in "
        + "the direction of the arrows.\nLeft and right arrows show indentation problems.";
    public static enum State { INTRO, START, UNJUMBLE, CHECK, QUESTION, OVER }
    private State state;
    private Grid grid = new Grid();
    private Button ctrlButton;
    private Button copyButton;
    private Button optionsButton;
    private Button hintButton;
    private Button helpButton;
    private Label problem;
    private Label instruct;
    private Label roundStatus;
    private Label consoleTitle;
    private Label gameStatus;
    private StopWatch stopwatch = new StopWatch();
    private int errorCount = 0;
    private long elapsedSeconds;
    private int score;
    private int priorScore;
    private int priorElapsedTime;
    private Label scoreLabel;
    private boolean dragEnabled;
    private Label gameOver;
    private Label scoreOver;
    private int matrixCounter;
    private String filename = "test.txt";
    // Game play
    private String problemText;
    private List<CodeLine> code;
    private ArrayList<Challenge> challenges;
    private int challengeNumber;
    private Challenge currChallenge;
    private String challengeCode = "";
    private String challengeTitle = "";
    private String challengeDescription = "";
    // Music and sounds
    private static GreenfootSound[] challengeMusicList = { 
            new GreenfootSound("challenge1.mid"), 
            new GreenfootSound("challenge2.mid"), 
            new GreenfootSound("challenge3.mid") 
        };
    private static GreenfootSound challengeMusic;
    private static GreenfootSound gameOverMusic = new GreenfootSound("gameover.mid");
    private static GreenfootSound gameOverSfx = new GreenfootSound("can-you-see-it-now.wav");

    /**
     * Constructor for Game world.
     */
    //public Game(ArrayList<Challenge> challenges) {    
    public Game(String fileName) {    
        super(WIDTH, HEIGHT, CELL_SIZE, false);
        LEFT_X = (int) (getWidth() / 4.0);
        RIGHT_X = (int) (getWidth() * 3.0 / 4.0);
        challengeNumber = 0;
        //this.challenges = challenges;
        try {
            loadChallenges(fileName);
        } catch(IOException ioe) {
            ioe.printStackTrace();
        }
        int randomMusic = Greenfoot.getRandomNumber(challengeMusicList.length);
        challengeMusic = challengeMusicList[randomMusic];
        showIntro();
        //gameOver();
    }

    /**
     * Coordinate animated aspects of each part of the scenario.
     */
    public void act() {
        if (state == State.UNJUMBLE) {
            setRoundStatus();
        } else if (state == State.CHECK) {
            //do check animations
        } else if (state == State.OVER) {
            gameOverAnim();
            matrixAnim();
            if (Greenfoot.mouseClicked(null)) {
                toMainMenu();
            }
        }
        if (state != State.OVER && !challengeMusic.isPlaying()
        && OptionsManager.INSTANCE.isEnabled(OptionsManager.MUSIC_ON_OFF)) {
            challengeMusic.play();
        }
        if (Greenfoot.isKeyDown("P")) Greenfoot.stop(); // for debugging
    }  

    /**
     * Responds to button events and controls the flow of the game.
     *
     * @param c The button or other GUIComponent that created the event.
     */
    public void actionPerformed(GUIComponent c) {
        // State control buttons
        if (c.getText().equalsIgnoreCase("Continue")) {
            gameStart();
        } else if (c.getText().equalsIgnoreCase("Start")) {
            unjumble();
        } else if (c.getText().equalsIgnoreCase("Check")) {
            checkValid();
        } else if (c.getText().equalsIgnoreCase("Next")) {
            //clearChallenge();
            ctrlButton.setText("Start");
            instruct.setText(INSTR_START);
            setupProblem();
        } else if (c.getText().equalsIgnoreCase("Over")) {
            gameOver();
        } else if (c.getText().equalsIgnoreCase("Main Menu")) {
            toMainMenu();
        }
        // Utility buttons
        if (c.getText().equalsIgnoreCase("Copy")) {
            copyCodeToTerminal();
        }
        if (c.getID() == 1000) { // help
            List objs = getObjects(Instructions.class);
            if (objs.size() > 0) {
                removeObjects(objs);
            } else {
                addObject(new Instructions(true), getWidth() / 2, getHeight() / 2);
            }
        }
        if (c.getID() == 1001) { // options
            OptionsManager opman = OptionsManager.INSTANCE;
            addObject(opman, WIDTH / 2, HEIGHT / 2);
        }
        if (c.getID() == 1002) { // code peek
            CodePanel cp = new CodePanel(currChallenge);
            addObject(cp, WIDTH / 2, HEIGHT / 2);
        }
    }

    // -- Utility
    public void addToScore(int amount) {
        score += amount;
        setGameStatus();
    }

    public boolean canDrag() {
        return dragEnabled;
    }

    public int getBottom() {
        return CODE_Y + code.size() - 1;
    }

    // for option manager
    public static void setChallengeMusicEnabled(boolean on) {
        if (on) {
            challengeMusic.play();
        } else {
            challengeMusic.stop();
        }
    }

    // -- Helper Methods
    private void toMainMenu() {
        List<Actor> list = getObjects(null);
        removeObjects(list);
        list.clear();
        list = null;
        challengeMusic.stop();
        gameOverMusic.stop();
        System.gc();
        //Greenfoot.setWorld(new GameManager(false));
        Greenfoot.setWorld(new GameManager());
    }

    private void showIntro() {
        state = State.INTRO;
        // NTR: do not like multiple paths
        if (challengeDescription.equals("")) {
            gameStart();
            return;
        }
        TextPanel panel = new TextPanel(challengeCode, challengeTitle, challengeDescription, this);
        addObject(panel, getWidth() / 2, getHeight() / 2); 
    }

    /**
     * Setup a challenge problem.
     */
    private void gameStart() {
        state = State.START;
        score = 0;
        List<Actor> list = getObjects(null);
        removeObjects(list);
        if (gameOver != null) {
            gameOver = null;
            scoreOver = null;
            //timeOver = null;
            //playAgain = null;
        }
        createConsole();
        setupProblem();
        elapsedSeconds = 0;
        System.gc();
    }

    private void unjumble() {
        state = State.UNJUMBLE;
        for (int i = 0; i < code.size(); i++) {
            addObject(code.get(i), RIGHT_X, CODE_Y + i);
        }
        problem.setEnabled(false); // Fade the instructions
        instruct.setText(INSTR_TEXT);
        stopwatch.start();
        dragEnabled = true;
        ctrlButton.setText("Check");
        grid.setEnabled(true);
    }

    private void setConsoleTitle() {
        String title = challengeTitle + ": " 
            + (challengeNumber + 1) + " / " + challenges.size();
        if (consoleTitle == null) {
            consoleTitle = new Label(title, MED_FONT, Color.BLUE);
        } else {
            consoleTitle.setText(title);
        }
    }

    private void createConsole() {
        drawBackground();
        addObject(grid, WIDTH / 4, CODE_Y + CODE_HEIGHT / 2);
        setConsoleTitle();
        addObject(consoleTitle, WIDTH / 2, 0);
        instruct = new Label(INSTR_START, MED_FONT, Color.BLUE, new Color(0, 0, 255, 36));
        instruct.setSize(new Dimension(500, 40));
        addObject(instruct, WIDTH / 2, CODE_Y - 3);
        addObject(new Label("Ordered Statements", MED_FONT, Color.BLUE), LEFT_X, CODE_Y - 1);
        Label codePile = new Label("Code Pile", MED_FONT, Color.BLUE);
        addObject(codePile, RIGHT_X, CODE_Y - 1);
        ctrlButton = new Button(BTN_UP, BTN_DWN, BTN_HVR);
        ctrlButton.setText("Start");
        ctrlButton.setFont(MED_FONT);
        ctrlButton.addActionListener(this);
        addObject(ctrlButton, LEFT_X + 4, CODE_Y + CODE_HEIGHT + 3);
        copyButton = new Button(BTN_UP, BTN_DWN, BTN_HVR);
        copyButton.setText("Copy");
        copyButton.setFont(MED_FONT);
        copyButton.addActionListener(this);
        helpButton = new Button(new GreenfootImage("help-up.gif"), 
            new GreenfootImage("help-dwn.gif"), new GreenfootImage("help-hvr.gif"));
        helpButton.setID(1000);
        helpButton.addActionListener(this);
        addObject(helpButton, WIDTH - 2, HEIGHT - 2);
        optionsButton = new Button(new GreenfootImage("settings_up.gif"), 
            new GreenfootImage("settings_dwn.gif"), new GreenfootImage("settings_hvr.gif"));
        optionsButton.setID(1001);
        optionsButton.addActionListener(this);
        addObject(optionsButton, WIDTH - 5, HEIGHT - 2);
        hintButton = new Button(new GreenfootImage("hint-up.gif"), 
            new GreenfootImage("hint-dwn.gif"), new GreenfootImage("hint-hvr.gif"));
        hintButton.setForeground(new Color(162, 234, 164));
        hintButton.setFont(PROB_FONT);
        hintButton.setText("Code\nPeek");
        hintButton.addActionListener(this);
        hintButton.setID(1002);
        addObject(hintButton, WIDTH - 8, HEIGHT - 2);
        roundStatus = new Label("");
        roundStatus.setSize(new Dimension(200, 20));
        gameStatus = new Label("");
        gameStatus.setSize(new Dimension(200, 20));
        setRoundStatus();
        setGameStatus();
        addObject(roundStatus, LEFT_X - 3, CODE_Y + CODE_HEIGHT + 3);
        addObject(gameStatus, RIGHT_X - 5, CODE_Y + CODE_HEIGHT + 3);
    }

    // called by OptionsManager
    public void drawBackground() {
        int centerX = WIDTH / 2 * CELL_SIZE + CELL_SIZE / 2;
        GreenfootImage bg = getBackground();
        bg.setColor(Color.WHITE);
        bg.fill();
        GameManager.drawWallpaper(bg);
        bg.setColor(Color.BLACK);
        bg.drawLine(centerX, CODE_Y * CELL_SIZE, centerX, (CODE_Y + CODE_HEIGHT + 2) * CELL_SIZE);

    }

    private void setRoundStatus() {
        String text = "" + errorCount + " error";
        if (errorCount != 1) text += "s";
        long seconds = stopwatch.getElapsedSeconds();
        text += ", " + seconds;
        text += " second";
        if (seconds != 1) text += "s";
        roundStatus.setText(text);
    }

    private void setGameStatus() {
        String text = "Score: " + score + "    Total seconds: " + elapsedSeconds;
        gameStatus.setText(text);
    }

    private void clearChallenge() {
        if (problem != null) removeObject(problem);
        if (code != null) removeObjects(code);
        if (copyButton != null) removeObject(copyButton);
        removeObjects(getObjects(Answer.class));
        removeObjects(getObjects(Question.class));
    }

    private void setupProblem() {
        clearChallenge();
        if (challenges == null || challenges.size() == 0) {
            gameOver();
            return;
        }
        setConsoleTitle();
        currChallenge = challenges.get(challengeNumber);
        problemText = currChallenge.getProblem();
        problem = new Label(problemText, PROB_FONT);
        problem.setHorizontalAlignment(Label.LEFT);
        problem.setSize(new Dimension((WIDTH - 2) * CELL_SIZE, 5 * CELL_SIZE));
        addObject(problem, getWidth() / 2, 3);
        code = currChallenge.getCodeLines();
        challengeNumber++;
        Collections.shuffle(code);
    }

    public void checkValid() {
        boolean valid = true;
        int numWrong = 0;
        for (int i = 0; i < code.size(); i++) {
            int x = LEFT_X - 1;
            int y = CODE_Y + i;
            // Make list of CodeLines at this line
            List<CodeLine> list = new ArrayList<CodeLine>();
            do {
                x++;
                List<CodeLine> list2 = getObjectsAt(x, y, CodeLine.class);
                for (CodeLine line : list2) {
                    if (!list.contains(line)) list.add(line);
                }
            } while (x < getWidth());
            // Check placement of lines
            for (int j = 0; j < list.size(); j++) {
                valid = list.get(j).checkLine(i + 1);
                if (!valid)  numWrong++;
            }
        }
        if (numWrong > 0) {
            valid = false;
            errorCount += numWrong;
        }
        // return valid;
        if (valid) {
            elapsedSeconds += stopwatch.getElapsedSeconds();
            stopwatch.stop();
            dragEnabled = false;
            score += currChallenge.getProblemValue() - errorCount;
            errorCount = 0;
            setGameStatus();
            Greenfoot.playSound("reward.wav");
            if (currChallenge.hasQuestion()) {
                addObject(currChallenge.getQuestion(), RIGHT_X, CODE_Y + CODE_HEIGHT / 2);
            }

            if (challengeNumber < challenges.size()) {
                ctrlButton.setText("Next");
                grid.setEnabled(false);
                if (currChallenge.hasQuestion()) {
                    instruct.setText(INSTR_NEXT_Q);
                } else {
                    instruct.setText(INSTR_NEXT);
                }
                if (OptionsManager.INSTANCE.isEnabled(OptionsManager.COPY_BUTTON)) {
                    addObject(copyButton, LEFT_X - 3, CODE_Y + CODE_HEIGHT + 3);
                }
            } else if (currChallenge.hasQuestion()) {
                // if valid and no question || valid and Next button
                ctrlButton.setText("Over");
            } else {
                gameOver();
            }
        } else {
            Greenfoot.playSound("uhoh.wav");
            instruct.setText(INSTR_ERROR);
        }
    }

    /**
     * Copy unjumbled code to the Terminal window.
     */
    private void copyCodeToTerminal() {
        System.out.println("Challenge " + challengeNumber + " code:");
        for (int i = 0; i < code.size(); i++) {
            int x = LEFT_X - 1;
            int y = CODE_Y + i;
            List<CodeLine> list = getObjectsAt(x, y, CodeLine.class);
            for (CodeLine line : list) {
                int indentLevel = line.getIndentationLevel();
                for (int indent = 0; indent < indentLevel; indent++) {
                    System.out.print("    ");
                }
                System.out.println(line.getText());
            }
        }
    }

    /**
     * Returns the maximum possible score for this set of challenges.
     *
     * @return The maximum possible score.
     */
    public int maxScore() {
        int sum = 0;
        for (Challenge c : challenges) {
            sum += c.getTotalValue();
        }
        return sum;
    }

    /**
     * Clear challenge screen and do gameOverAnim() until restarted.
     */
    public void gameOver() {
        state = State.OVER;
        List<Actor> list = getObjects(null);
        removeObjects(list);
        GreenfootImage bg = getBackground();
        bg.setColor(Color.BLACK);
        bg.fill();
        Font f = bg.getFont().deriveFont(52f);
        Color c = new Color(0, 255, 0, 0);
        gameOver = new Label("Game Over", f, c);
        addObject(gameOver, getWidth() / 2, getHeight() / 3);
        addObject(new Symbol("GAME", true), getWidth() / 3, -1);
        addObject(new Symbol("OVER", true), getWidth() / 3 * 2, -1);
        saveScore();
        challengeMusic.stop();
        if (OptionsManager.INSTANCE.isEnabled(OptionsManager.MUSIC_ON_OFF)) {
            gameOverMusic.play();
        }
    }

    public void saveScore() {
        priorElapsedTime = Leaderboard.INSTANCE.getElapsedTime(challengeCode);
        priorScore = Leaderboard.INSTANCE.updateScore(challengeCode, score, elapsedSeconds, maxScore());
    }

    private void gameOverAnim() {
        Color c = gameOver.getForeground();
        int alpha = c.getAlpha();
        if (alpha < 255) {
            alpha++;
            gameOver.setForeground(new Color(c.getRed(), c.getGreen(), c.getBlue(), alpha));
        } else if (scoreOver == null) {
            Font f = gameOver.getFont().deriveFont(36f);
            int x = getWidth() / 2;
            int yStart = getHeight() / 2 - 1;
            int y = yStart;
            int maxScore = maxScore();
            String scoreMsg = "Score: " + score + " / " + maxScore;
            scoreOver = new Label(scoreMsg, f, Color.GREEN); 
            addObject(scoreOver, x, y);
            y += 2;
            String msg = "Total seconds: " + elapsedSeconds;
            addObject(new Label(msg, f, Color.GREEN), x, y);
            gameOverSfx.play();
            if (!UserInfo.isStorageAvailable()) {
                msg = "Login to save scores and time";
                addObject(new Label(msg, f, Color.GREEN), x, yStart + 7);
            } else {
                //Font f2 = f.deriveFont(20f);
                y += 2;
                msg = "Prior best score: ";
                msg += (priorScore > -1) ? priorScore : "none";
                addObject(new Label(msg, f, Color.GREEN), x, y);
                y += 2;
                msg = "Prior elapsed time: ";
                msg += (priorElapsedTime > -1) ? priorElapsedTime : "none";
                addObject(new Label(msg, f, Color.GREEN), x, y);
                if (score > priorScore && score == maxScore) {
                    y += 2;
                    msg = "Reward: \u2605\u2605\u2605";
                    addObject(new Label(msg, f, Color.GREEN), x, y);
                    RewardsStore.INSTANCE.addReward(3);
                    RewardsStore.INSTANCE.addTotalReward(3);
                }
            }
            y += 3;
            addObject(new Label("Click to play again", f, Color.GREEN), x, y);

        }
    }

    private void matrixAnim() {
        matrixCounter++;
        if (matrixCounter > 9) {
            int x = Greenfoot.getRandomNumber(WIDTH);
            if (code == null && challenges.size() > 0) {
                currChallenge = challenges.get(challengeNumber);
                code = currChallenge.getCodeLines();
            }
            if (code == null || code.size() <= 0) return;
            int rand = Greenfoot.getRandomNumber(code.size());
            String str = code.get(rand).getText();
            addObject(new Symbol(str, true), x , -1);
            matrixCounter = 0;
        }
    }

    // -- File Loading methods --
    public void loadChallenges(String filename) throws IOException {
        challenges = new ArrayList<Challenge>();
        //challengeNumber = 0;
        filename = "challenges/" + filename;
        ArrayList<String> lines = loadFile(filename);
        filterUnicode(lines);
        //filterNewLines(lines);
        parseLines(lines);
    }

    private ArrayList<String> loadFile(String filename) throws IOException {
        ArrayList<String> lines = new ArrayList<String>();
        // Allow loading from JAR files and Greenfoot Gallery
        ClassLoader classLoader = getClass().getClassLoader();
        InputStream is = classLoader.getResourceAsStream(filename);
        if (is == null) {
            throw new IOException("No such file: " + filename);
        }
        BufferedReader reader = new BufferedReader(
                new InputStreamReader(is));
        String line = reader.readLine();
        while (line != null) {
            if (!line.equals("") && !line.startsWith("#")) {
                lines.add(line);
            }
            line = reader.readLine();
        }
        reader.close();
        return lines;
    }

    private void filterUnicode(ArrayList<String> lines) {
        Pattern unicode = Pattern.compile("\\\\u([0-9A-Fa-f]+)");
        for (int i = 0; i < lines.size(); i++) {
            Matcher m = unicode.matcher(lines.get(i));
            StringBuffer sb = new StringBuffer();
            while (m.find()) {
                int code = Integer.valueOf(m.group(1), 16);
                String uc = "" + Character.valueOf((char)code);
                m.appendReplacement(sb, uc);
            }
            m.appendTail(sb);
            lines.set(i, new String(sb));
        }
    }

    private void filterNewLines(ArrayList<String> lines) {
        //Pattern newline = Pattern.compile("\\\\n");
        for (int i = 0; i < lines.size(); i++) {
            Matcher m = newline.matcher(lines.get(i));
            StringBuffer sb = new StringBuffer();
            while (m.find()) {
                m.appendReplacement(sb, "\n");
            }
            m.appendTail(sb);
            lines.set(i, new String(sb));
        }
    }

    private Pattern newline = Pattern.compile("\\\\n");
    private String filterNewLines(String line) {
        Matcher m = newline.matcher(line);
        StringBuffer sb = new StringBuffer();
        while (m.find()) {
            m.appendReplacement(sb, "\n");
        }
        m.appendTail(sb);
        return new String(sb);
    }

    private void parseLines(final ArrayList<String> lines) {
        Matcher meta = Pattern.compile("(?is):META:(\\w+):(.+)").matcher("");
        Matcher desc = Pattern.compile("(?is):DESCRIPTION:(.+)").matcher("");
        Matcher prob = Pattern.compile("(?is):PROB:(\\d+):(.+)").matcher("");
        Matcher code = Pattern.compile("(?i):CODE:(\\d+):(\\d+):(\\d+):(.+)").matcher("");
        Matcher qstn = Pattern.compile("(?is):QUEST:(\\d+):(.+)").matcher("");
        Matcher answ = Pattern.compile("(?i):ANSWER:([TF]):(.+)").matcher("");
        Matcher ordr = Pattern.compile("(?i):ANSWERORDER:(.+)").matcher("");
        Matcher expl = Pattern.compile("(?is):EXPLANATION:(.+)").matcher("");
        int challengeNum = -1;
        Challenge c = null;
        for (int i = 0; i < lines.size(); i++) {
            String line = lines.get(i);
            //System.out.println(line);
            //prob.reset(line);
            //code.reset(line);
            //qstn.reset(line);
            //answ.reset(line);
            //ordr.reset(line);
            //expl.reset(line);
            // if (prob.matches()) {
            try {
                if (line.startsWith(":PROB:")) {
                    line = filterNewLines(line);
                    prob.reset(line);
                    prob.matches();
                    c = new Challenge(prob.group(2), Integer.valueOf(prob.group(1)));
                    challenges.add(c);
                    //} else if (code.matches()) {
                } else if (line.startsWith(":CODE:")) {
                    //System.out.println("matching code: " + line);
                    code.reset(line);
                    code.matches();
                    int indent = Integer.valueOf(code.group(1));
                    int start = Integer.valueOf(code.group(2));
                    int end = Integer.valueOf(code.group(3));
                    c.addCodeLine(new CodeLine(code.group(4), indent, start, end));
                    //} else if (qstn.matches()) {
                } else if (line.startsWith(":QUEST:")) {
                    line = filterNewLines(line);
                    qstn.reset(line);
                    qstn.matches();
                    c.setQuestionText(qstn.group(2));
                    c.setQuestionValue(Integer.valueOf(qstn.group(1)));
                    //} else if (answ.matches()) {
                } else if (line.startsWith(":ANSWER:")) {
                    answ.reset(line);
                    answ.matches();
                    String correct = answ.group(1);
                    if (correct.equalsIgnoreCase("T")) {
                        c.addAnswer(answ.group(2), true);
                    } else {
                        c.addAnswer(answ.group(2), false);
                    }
                    //} else if (ordr.matches()) {
                } else if (line.startsWith(":ANSWERORDER:")) {
                    ordr.reset(line);
                    ordr.matches();
                    if (ordr.group(1).equalsIgnoreCase("randomized")) {
                        c.shuffleAnswers();
                    }
                    //} else if (expl.matches()) {
                } else if (line.startsWith(":EXPLANATION:")) {
                    line = filterNewLines(line);
                    expl.reset(line);
                    expl.matches();
                    c.setExplanation(expl.group(1));
                } else if (line.startsWith(":META:")) {
                    meta.reset(line);
                    meta.matches();
                    challengeCode = meta.group(1);
                    challengeTitle = meta.group(2);
                    //System.out.println(challengeCode+":"+description);
                } else if (line.startsWith(":DESCRIPTION:")) {
                    line = filterNewLines(line);
                    desc.reset(line);
                    desc.matches();
                    challengeDescription = desc.group(1);
                } else {
                    System.out.println("Unrecognized line: " + line);
                }
            } catch(IllegalStateException ise) {
                System.out.println(ise.getMessage() + " for line: " + line);
            }
        }
    }
}
