import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;

/**
 * A leaderboard for Jumbled games.
 * 
 * @author Ed Parrish
 * @version 1.0
 */
public class Leaderboard extends Actor {
    public static final Leaderboard INSTANCE = new Leaderboard();
    private static final int SUM_REWARD_INT = RewardsStore.SUM_REWARD_INT;
    private static final int ENTRY_LENGTH = 5;
    private static final int SCORE_CHAR = ENTRY_LENGTH - 2;
    private static final int TIME_CHAR = ENTRY_LENGTH - 1;
    private static final int WIDTH = (GameManager.WIDTH - 20);// * GameManager.CELL_SIZE;
    private static final int HEIGHT = (GameManager.HEIGHT - 20);// * GameManager.CELL_SIZE;
    private static GreenfootImage bg = new GreenfootImage(WIDTH, HEIGHT);
    private static final Font TITLE_FONT = new Font("SansSerif", Font.BOLD, 16);
    private static final Font TEXT_FONT = new Font("SansSerif", Font.PLAIN, 14);
    private static final Color TEXT_COLOR = Color.BLACK;
    private static final Color BG_COLOR = Color.WHITE;
    private static final Color OUTLINE_COLOR = Color.BLUE;
    private static final int RANK_X = 20;
    private static final int SCORE_X = 50;
    private static final int NAME_X = 110;
    private static final int BADGE_X = 220;
    private static final GreenfootImage badge = new GreenfootImage("badge-smiley.png");

    /**
     * Leaderboard constructor is private for singleton pattern.
     */
    private Leaderboard() { 
        // Exists only to defeat instantiation.
    }

    /**
     * Respond to mouse clicks on the background. 
     */
    public void act() {
        if (Greenfoot.mouseClicked(this)) {
            getWorld().removeObject(this);
        }
    }

    public void showTopScores(int num) {
        drawBackground();
        bg.setColor(TEXT_COLOR);
        bg.setFont(TITLE_FONT);
        UserInfo myInfo = GameManager.getUserInfo();
        if (myInfo == null) {
            bg.drawString("You must be logged in to view the leaderboard", RANK_X, 40);
        } else {
            String title = "Jumbled Games Leaderboard";
            Graphics g = getImage().getAwtImage().getGraphics();
            Font font = bg.getFont();
            FontMetrics metrics = g.getFontMetrics(font);
            int adv = metrics.stringWidth(title);
            bg.drawString(title, (WIDTH - adv) / 2, 30);
            // Draw subtitle
            bg.setFont(TEXT_FONT);
            //UserInfo myInfo = GameManager.getUserInfo();
            String info = myInfo.getUserName() + " level " + rewardedLevels(myInfo)
                + "  \u2022  Score: " + myInfo.getScore();
            int y = 50;
            metrics = g.getFontMetrics(TEXT_FONT);
            adv = metrics.stringWidth(info);
            bg.drawString(info, (WIDTH - adv) / 2, y);
            // Draw the top scores
            y = 70;
            for (Object obj : UserInfo.getTop(num)) {
                UserInfo user = (UserInfo) obj;  
                String username = user.getUserName(); 
                int score = user.getScore();  
                GreenfootImage userImg = user.getUserImage();
                bg.drawImage(userImg, 20, y - 12);
                bg.drawString("" + username, 80, y);
                int levels = completedLevels(user);
                int rewardedLevels = rewardedLevels(user);
                String msg = "Rewarded levels: " + rewardedLevels
                    + "  \u2022  Attempted levels: " + levels 
                    + "  \u2022  Score: " + score;
                bg.drawString(msg, 80, y + 16);
                for (int i = 0; i < rewardedLevels; i++) {
                    bg.drawImage(badge, 80 + i * 20, y + 24);
                }
                y += 70;
            }
            bg.drawString("Click on leaderboard background to return", RANK_X, HEIGHT - 16);
        }
        setImage(bg);
    }

    /**
     * Returns the number of completed levels for the current player.
     *
     * @return The number of completed levels of -1 if not logged in.
     */
    public int completedLevels() {
        UserInfo currentUser = GameManager.getUserInfo();
        if (currentUser == null) {
            return -1;
        }
        return completedLevels(currentUser);
    }

    private int completedLevels(UserInfo user) {
        StringBuilder sb = readStrings(user);
        return sb.length() / ENTRY_LENGTH;
    }
    
    /**
     * Returns the number of levels for which rewards were given to the current player.
     *
     * @return The number of rewarded levels.
     */
    public int rewardedLevels() {
        UserInfo currentUser = GameManager.getUserInfo();
        if (currentUser == null) {
            return -1;
        }
        return rewardedLevels(currentUser);
    }

    // precondition: user cannot be null
    private int rewardedLevels(UserInfo user) {
        final int REWARDS_PER_LEVEL = 3;
        int sumRewards = user.getInt(SUM_REWARD_INT);
        return sumRewards / REWARDS_PER_LEVEL;
    }

    /**
     * Returns the current score for a specified challenge code.
     *
     * @param challengeCode The three-digit challenge code.
     * @return The score, or -1 if the score is not available.
     */
    public int getScore(String challengeCode) {
        int currentScore = -1; // invalid score
        if (UserInfo.isStorageAvailable()) {
            UserInfo myInfo = UserInfo.getMyInfo();
            StringBuilder sb = readStrings(myInfo);
            int index = findIndex(sb, challengeCode);
            if (index >= 0) {
                currentScore = (int) sb.charAt(index + SCORE_CHAR);
            }
        }
        return currentScore;
    }

    /**
     * Returns the current elapsed time for a specified challenge code.
     *
     * @param challengeCode The three-digit challenge code.
     * @return The current elapsed time, or -1 if the time is not available.
     */
    public int getElapsedTime(String challengeCode) {
        int currentTime = -1; // invalid time
        if (UserInfo.isStorageAvailable()) {
            UserInfo myInfo = UserInfo.getMyInfo();
            StringBuilder sb = readStrings(myInfo);
            int index = findIndex(sb, challengeCode);
            if (index > -1) {
                currentTime = (int) sb.charAt(index + TIME_CHAR);
            }
        }
        return currentTime;
    }

    /**
     * Updates the score and time for the challengeCode if the score is higher
     * or the time if the score is the same or better and the time is lower.
     *
     * @param challengeCode The three digit challenge code.
     * @param score The new score.
     * @param seconds The time for the new score.
     * @param maxScore The maximum score for a challenge game.
     * @return The prior score, or -1 if the score is not available.
     */
    public int updateScore(String challengeCode, int score, long seconds, int maxScore) {
        int oldScore = -1; // invalid score
        if (UserInfo.isStorageAvailable()) {
            UserInfo myInfo = UserInfo.getMyInfo();
            StringBuilder sb = readStrings(myInfo);
            int index = findIndex(sb, challengeCode);
            boolean updated = false;
            if (index < 0) { // new entry
                sb.append(challengeCode);
                sb.append((char) score);
                sb.append((char) seconds);
                int sumScores = totalScore(sb);
                myInfo.setScore(sumScores);
                updated = true;
            } else { // update existing score
                oldScore = (int) sb.charAt(index + SCORE_CHAR);
                if (oldScore > maxScore) { // in case points are reduced
                    oldScore = maxScore;
                    updated = true;
                }
                int oldTime = (int) sb.charAt(index + TIME_CHAR);
                if (score > oldScore || updated) {
                    sb.setCharAt(index + SCORE_CHAR, (char) score);
                    sb.setCharAt(index + TIME_CHAR, (char) seconds);
                    int sumScores = totalScore(sb);
                    myInfo.setScore(sumScores);
                    updated = true;
                } else if (score == oldScore && oldTime > seconds) {
                    sb.setCharAt(index + TIME_CHAR, (char) seconds);
                    updated = true;
                }
            }
            if (updated) { // save only if updated to reduce server load
                writeStrings(myInfo, sb);
                myInfo.store();
            }
        }
        return oldScore;
    }

    private void drawBackground() {
        bg.setColor(OUTLINE_COLOR);
        bg.fill();
        bg.setColor(BG_COLOR);
        bg.fillRect(2, 2, WIDTH - 4, HEIGHT - 4);
        GameManager.drawWallpaper(bg);
    }

    private int totalScore(StringBuilder info) {
        int sum = 0;
        for (int i = 0; i < info.length(); i += ENTRY_LENGTH) {
            sum += (int) info.charAt(i + SCORE_CHAR);
        }
        return sum;
    }

    private int findIndex(StringBuilder info, String challengeCode) {
        for (int i = 0; i < info.length(); i += ENTRY_LENGTH) {
            // NTR: corrupted data could throw StringIndexOutOfBoundsException
            String entry = info.substring(i, i + ENTRY_LENGTH);
            if (entry.startsWith(challengeCode)) return i;
        }
        return -1;
    }

    private StringBuilder readStrings(UserInfo ui) {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < UserInfo.NUM_STRINGS; i++) {
            sb.append(ui.getString(i));
        }
        return sb;
    }

    private void writeStrings(UserInfo ui, StringBuilder sb) {
        int len = UserInfo.STRING_LENGTH_LIMIT;
        int numStrings =  sb.length() / len + 1;
        if (numStrings > UserInfo.NUM_STRINGS) {
            numStrings = UserInfo.NUM_STRINGS;
        }
        for (int i = 0; i < numStrings; i++) {
            int end = (i + 1) * len;
            if (end > sb.length()) end = sb.length();
            String str = sb.substring(i * len, end);
            ui.setString(i, str);
        }
        ui.store();
    }
}
