import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.util.ArrayList;
import java.util.List;

/**
 * Manages the options in the game.
 *
 * @author Ed Parrish
 * @version 7/15/2014
 */
public class OptionsManager extends Actor implements ActionListener
{
    /**
     * RewardsStore instance for singleton pattern.
     */
    public static final OptionsManager INSTANCE = new OptionsManager();
    public static final int COPY_BUTTON = RewardsStore.COPY_BUTTON;
    public static final int CODE_PEEK = RewardsStore.CODE_PEEK;
    public static final int MUSIC_ON_OFF = 30;
    public static final int SFX_ON_OFF = 29;
    private static final GreenfootImage BTN_UP = new GreenfootImage("roundbtn-up.gif");
    private static final GreenfootImage BTN_DWN = new GreenfootImage("roundbtn-down.gif");
    private static final GreenfootImage BTN_HVR = new GreenfootImage("roundbtn-hover.gif");
    private static final int WIDTH = (GameManager.WIDTH - 10);
    private static final int HEIGHT = (GameManager.HEIGHT - 10);
    private static GreenfootImage bg = new GreenfootImage(WIDTH, HEIGHT);
    private static final Color TITLE_COLOR = Color.BLUE; //new Color(212, 175, 55);
    private static final Color TEXT_COLOR = Color.BLACK;
    private static final Color BG_COLOR = Color.WHITE;
    private static final Color OUTLINE_COLOR = Color.BLUE;
    private static final Font TITLE_FONT = new Font("SansSerif", Font.BOLD, 16);
    private static final Font HEADING_FONT = new Font("SansSerif", Font.BOLD, 14);
    private static final Font TEXT_FONT = new Font("SansSerif", Font.PLAIN, 14);
    // Following storage ints must not conflict with RewardsStore
    private static final int SETTINGS_INT = 0; // record of settings
    private static final int TOKENS_INT = RewardsStore.TOKENS_INT; // number of tokens purchased
    private ArrayList<Button> buttonList = new ArrayList<Button>();
    private ArrayList<Setting> options = new ArrayList<Setting>();

    /**
     * OptionManager constructor is private for singleton pattern.
     */
    private OptionsManager() {
        buildOptions(); // can be accessed before opening panel
    }

    /**
     * Respond to mouse clicks on the background.
     */
    public void act() {
        if (Greenfoot.mouseClicked(this)) {
            World w = getWorld();
            w.removeObjects(buttonList);
            buttonList.clear();
            w.removeObject(this);
            System.gc();
        }
    }

    /**
     * Responds to button events and controls the flow of the game.
     *
     * @param c The button or other GUIComponent that created the event.
     */
    public void actionPerformed(GUIComponent c) {
        if (c.getText().equals("On") && !isEnabled(c.getID())) {
            setOption(c.getID());
            updateComponents(c);
            if (c.getID() == MUSIC_ON_OFF) {
                World w = getWorld();
                if (w instanceof GameManager) {
                    GameManager.setMusicEnabled(true);
                } else if (w instanceof Game) {
                    Game.setChallengeMusicEnabled(true);
                }
            }
        } else if (c.getText().equals("Off") && isEnabled(c.getID())) {
            clearOption(c.getID());
            updateComponents(c);
            if (c.getID() == MUSIC_ON_OFF) {
                World w = getWorld();
                if (w instanceof GameManager) {
                    GameManager.setMusicEnabled(false);
                } else if (w instanceof Game) {
                    Game.setChallengeMusicEnabled(false);
                }
            }
        }
        if (isWallpaper(c.getID())) {
            World w = getWorld();
            if (w instanceof GameManager) {
                GameManager gm = (GameManager) w;
                gm.drawBackground();
            } else if (w instanceof Game) {
                Game g = (Game) w;
                g.drawBackground();
            }
        }
        redrawPanel();
    }

    private void updateComponents(GUIComponent c) {
        UserInfo info = GameManager.getUserInfo();
        for (Button b: buttonList) {
            if (b != c && b.getID() == c.getID()) {
                b.setEnabled(true);
                c.setEnabled(false);
            }
        }
    }

    /**
     * Draws the panel after being added to the world.
     *
     * @param world The world to which the object was added.
     */
    protected void addedToWorld(World world) {
        buildOptions();
        drawPanel();
    }

    /**
     * Returns
     *
     * @return The return value
     */
    public List<OptionsManager.Setting> getWallpaper() {
        ArrayList<Setting> wallpaper = new ArrayList<Setting>();
        for (Setting set: options) {
            String name = set.getName().toLowerCase();
            if (isEnabled(set.id) && name.contains("wallpaper")) {
                wallpaper.add(set);
            }
        }
        return wallpaper;
    }

    private boolean isWallpaper(int id) {
        for (Setting set: options) {
            String name = set.getName().toLowerCase();
            if (id == set.id && name.contains("wallpaper")) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns whether a setting with a particular index has been set.
     *
     * @param id The id of the product.
     * @return Returns true if UserInfo storage is available and the product
     * given by the index has been set; otherwise returns false.
     */
    public boolean isEnabled(int id) {
        boolean selected = false;
        UserInfo myInfo = GameManager.getUserInfo();
        if (myInfo != null) {
            int settings = myInfo.getInt(SETTINGS_INT);
            selected = (settings & (1 << id)) != 0;
        }
        return selected;
    }

    // http://stackoverflow.com/questions/4674006/set-specific-bit-in-byte
    // set: myByte |= 1 << bit;
    public void setOption(int id) {
        UserInfo myInfo = GameManager.getUserInfo();
        if (myInfo != null) {
            Setting set = getSettingFromID(id);
            int settings = myInfo.getInt(SETTINGS_INT);
            settings |= 1 << id;
            myInfo.setInt(SETTINGS_INT, settings);
            myInfo.store();
        }
    }

    // clear: myByte &= ~(1 << bit);
    public void clearOption(int id) {
        UserInfo myInfo = GameManager.getUserInfo();
        if (myInfo != null) {
            Setting p = getSettingFromID(id);
            int settings = myInfo.getInt(SETTINGS_INT);
            settings &= ~(1 << id);
            myInfo.setInt(SETTINGS_INT, settings);
            myInfo.store();
        }
    }

    private Setting getSettingFromID(int id) {
        for (Setting set: options) {
            if (set.id == id) {
                return set;
            }
        }
        return null;
    }

    private void buildOptions() {
        options.clear();
        options.add(new Setting(MUSIC_ON_OFF, "notes.gif", "Background music"));
        //options.add(new Setting(SFX_ON_OFF, "default.gif", "Sound effects"));
        RewardsStore store = RewardsStore.INSTANCE;
        List<RewardsStore.Product> prods = store.getPurchasedProducts();
        for (RewardsStore.Product p: prods) {
            if (p.getId() != CODE_PEEK) {
                options.add(new Setting(p.getId(), p.getImageFile(), p.getName()));
            }
        }
    }

    private void redrawPanel() {
        World w = getWorld();
        w.removeObjects(buttonList);
        buttonList.clear();
        drawPanel();
    }

    private void drawPanel() {
        drawBackground();
        if (!UserInfo.isStorageAvailable()) {
            bg.setFont(TITLE_FONT);
            bg.setColor(TEXT_COLOR);
            bg.drawString("You must be logged in to use this feature", 20, 40);
            setImage(bg);
            return;
        }
        int y = drawTitle(bg);
        listOptions(bg, y + 24);
        setImage(bg);
    }

    private void drawBackground() {
        bg.setColor(OUTLINE_COLOR);
        bg.fill();
        bg.setColor(BG_COLOR);
        bg.fillRect(2, 2, WIDTH - 4, HEIGHT - 4);
        GameManager.drawWallpaper(bg);
        int x = 10;
        int y = bg.getHeight() - 8;
        bg.setColor(Color.DARK_GRAY);
        bg.setFont(TEXT_FONT);
        bg.drawString("Click on background to return", x, y);
    }

    private int drawTitle(GreenfootImage img) {
        img.setFont(TITLE_FONT);
        img.setColor(TITLE_COLOR);
        String title = "Jumbled Games Options and Settings";
        Graphics g = getImage().getAwtImage().getGraphics();
        FontMetrics metrics = g.getFontMetrics(TITLE_FONT);
        int adv = metrics.stringWidth(title);
        int y = metrics.getHeight() * 3 / 2;
        img.drawString(title, (WIDTH - adv) / 2, y);
        // draw name as subtitle
        int lineHeight = metrics.getHeight();
        y += (int)(lineHeight * 1.2);
        img.setFont(HEADING_FONT);
        img.setColor(TEXT_COLOR);
        UserInfo myInfo = GameManager.getUserInfo();
        String userName = myInfo.getUserName();
        metrics = g.getFontMetrics(HEADING_FONT);
        adv = metrics.stringWidth(userName);
        img.drawString(userName, (WIDTH - adv) / 2, y);
        return y;
    }

    private void listOptions(GreenfootImage bg, int startY) {
        Graphics g = bg.getAwtImage().getGraphics();
        g.setFont(TEXT_FONT);
        bg.setColor(TEXT_COLOR);
        FontMetrics metrics = g.getFontMetrics();
        int asc = metrics.getAscent();
        int y = startY;
        // draw headings
        bg.setFont(HEADING_FONT);
        bg.drawString("Icon", 20, y);
        bg.drawString("Option Name", 80, y);
        bg.drawString("State", 240, y);
        bg.drawString("Change To", 300, y);
        bg.setFont(TEXT_FONT);
        y += 25;
        UserInfo myInfo = GameManager.getUserInfo();
        for (Setting set: options) {
            GreenfootImage img = null;
            if (set.name.startsWith("Avatar") && myInfo != null) {
                img = myInfo.getUserImage();
            } else {
                img = new GreenfootImage(set.imageFile);
            }
            bg.drawImage(img, 20, y - asc);
            drawWrappedText(set.getName(), 80, y, 150, g);
            String state = "Off";
            if (isEnabled(set.id)) {
                state = "On";
            }
            bg.drawString(state, 240, y);
            placeButton(set, "On", 300, y);
            placeButton(set, "Off", 300 + 60, y);
            y += 60;
        }
    }

    private void drawWrappedText(String text, int x, int y, int width, Graphics g) {
        String[] lines = GUIComponent.wordWrap(text, width, g);
        FontMetrics metrics = g.getFontMetrics();
        int dy = metrics.getHeight();
        for (String str: lines) {
            bg.drawString(str, x, y);
            y += dy;
        }
    }

    private void placeButton(Setting set, String text, int x, int y) {
        boolean selected = isEnabled(set.getId());
        Button b = new Button(BTN_UP, BTN_DWN, BTN_HVR);
        b.setID(set.getId());
        b.setFont(TEXT_FONT);
        b.addActionListener(this);
        if (text.equals("On")) {
            b.setText(text);
            if (selected) b.setEnabled(false);
        } else if (text.equals("Off")) {
            b.setText(text);
            if (!selected) b.setEnabled(false);
        }
        buttonList.add(b);
        int cellSize = getWorld().getCellSize();
        int buttonX = x / cellSize + b.getWidth() / cellSize / 2;
        int buttonY = y / cellSize + b.getHeight() / cellSize / 2;
        getWorld().addObject(b, buttonX, buttonY);
    }

    public static class Setting {
        private int id;
        private String imageFile;
        private String name;

        Setting(int index, String imgFile, String name) {
            this.id = index;
            this.imageFile = imgFile;
            this.name = name;
        }

        public int getId() {
            return id;
        }

        public String getName() {
            return name;
        }
    }
}
