import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.util.ArrayList;
import java.util.List;

/**
 * Create particles from text images for special effects.
 *
 * @author Ed Parrish
 * @version 1.0 7/6/2014
 */
public class ParticleText extends Actor
{
    private static final int DEFAULT_NUM = 200;
    private GreenfootImage textImage;
    private List<Particle> particles;
    private int numParticles = DEFAULT_NUM;
    private int width; // width of the image in pixels.
    private int height; // height of the image in pixels.
    private int leftX; // leftmost x-coordinate of the image in pixels.
    private int upperY; // uppermost y-coordinate of the image in pixels.
    private int count;

    /**
     * Display default particle text.
     */
    public ParticleText() {
        this("Particle Text");
    }

    /**
     * Display the specified text with a default size, color and background.
     *
     * @param text The text to display.
     */
    public ParticleText(String text) {
        this(text, 64, null, null, DEFAULT_NUM);
    }

    /**
     * Display the text with the specified size, color and background, using a
     * default number of particles.
     *
     * @param text The text to display.
     * @param size The size of each line of text in pixels.
     * @param foreground The color of the text, or null for black.
     * @param background The color of the background image behind the text, or null
     *                   for a transparent background.
     */
    public ParticleText(String text, int size, Color foreground, Color background) {
        this(text, size, foreground, background, DEFAULT_NUM);
    }

    /**
     * Display the text with the specified size, color, background and number of
     * particles.
     *
     * @param text The text to display.
     * @param size The size of each line of text in pixels.
     * @param foreground The color of the text, or null for black.
     * @param background The color of the background image behind the text, or null
     *                   for a transparent background.
     * @param numParticles The number of particles to use displaying the text.
     */
    public ParticleText(String text, int size, Color fg, Color bg, int numParticles) {
        this.numParticles = numParticles;
        textImage = new GreenfootImage(text, size, fg, bg);
        width = textImage.getWidth();
        height = textImage.getHeight();
        setImage(textImage);
        particles = new ArrayList<Particle>();
    }

    /**
     * Display the text image with the specified number of particles.
     *
     * @param image The image to display.
     * @param numParticles The number of particles to use displaying the text.
     */
    public ParticleText(GreenfootImage image, int numParticles) {
        this.numParticles = numParticles;
        textImage = image;
        width = textImage.getWidth();
        height = textImage.getHeight();
        setImage(textImage);
        particles = new ArrayList<Particle>();
    }

    /**
     * Create the particles and disperse the text.
     *
     * @param world The world the particles live within.
     */
    protected void addedToWorld(World world) {
        leftX = getX() - width / 2;
        upperY = getY() - height / 2;
        double pixelsPerParticle = (double) width * height / numParticles;
        if (pixelsPerParticle < 1) {
            pixelsPerParticle = 1;
        }
        int pSize = (int)Math.round(Math.sqrt(pixelsPerParticle));
        makeParticles(pSize, pSize);
        for (Particle p: particles) {
            world.addObject(p, p.targetX, p.targetY);
        }
        setImage(new GreenfootImage(1, 1));
        disperse();
    }

    /**
     * Control the movement of particles.
     */
    public void act() {
        if (!isCoalesced()) {
            coalesce(2);
        }
    }

    /**
     * Move the image particles towards their target coordinates.
     *
     * @param speed The distance to move in pixels.
     */
    public void coalesce(int speed) {
        count++;
        for (Particle p: particles) {
            if (count < 256) {
                p.setTransparency(count);
            }
            p.moveTowardsTarget(speed);
        }
    }

    /**
     * Tests if all text particles have reached their target coordinates.
     *
     * @return Returns true of all text particles have reached their target
     * coordinates; otherwise returns false.
     */
    public boolean isCoalesced() {
        for (Particle p: particles) {
            if (!p.isAtTarget()) {
                return false;
            }
        }
        return true;
    }

    /**
     * Randomly scatters the text particles within the world.
     */
    public void disperse() {
        World w = getWorld();
        int width = w.getWidth();
        int height = w.getHeight();
        for (Particle p: particles) {
            int x = Greenfoot.getRandomNumber(width);
            int y = Greenfoot.getRandomNumber(height);
            p.setLocation(x, y);
            p.setTransparency(0);
        }
    }

    /**
     * Returns the number of particles in the particle text.
     *
     * @return The number of particles.
     */
    public int numberOfParticles() {
        return particles.size();
    }

    /**
     * Deletes all the particles in the particle text.
     */
    public void deleteParticles() {
        World w = getWorld();
        w.removeObjects(particles);
        particles.clear();
    }

    /**
     * Make particles from the image.
     *
     * @param pWidth The width of each particle.
     * @param pHeight The height of each particle.
     */
    private void makeParticles(int pWidth, int pHeight) {
        for(int x = 0; x < width; x += pWidth) {
            for(int y = 0; y < height; y += pHeight) {
                makeSingleParticle(x, y, pWidth, pHeight);
            }
        }
    }

    /**
     * Make a single particle of the specified width and height starting at the
     * upper left-hand pixel specified by the (x, y) coordinates.
     *
     * @param x The x-coordinate of the particle within the image.
     * @param y The y-coordinate of the particle within the image.
     * @param pWidth The width of the particle.
     * @param pHeight The height of the particle.
     */
    private void makeSingleParticle(int x, int y, int pWidth, int pHeight) {
        GreenfootImage pImg = new GreenfootImage(pWidth, pHeight);
        for (int i = 0; i < pWidth; i++) {
            for (int j = 0; j < pHeight; j++) {
                // Make certain the particle does not exceed image bounds.
                if (x + i < width && y + j < height) {
                    Color pixel = textImage.getColorAt(x + i, y + j);
                    pImg.setColorAt(i, j, pixel);
                }
            }
        }
        int centerX = leftX + x + pWidth / 2;
        int centerY = upperY + y + pHeight / 2;
        particles.add(new Particle(pImg, centerX, centerY));
    }
    
    /**
     * A simple particle containing part of the text image.
     *
     * @author Ed Parrish
     * @version 1.0  7/06/2014
     */
    private class Particle extends Actor {
        int targetX, targetY;
        GreenfootImage image;

        public Particle(GreenfootImage img, int targetX, int targetY) {
            this.targetX = targetX;
            this.targetY = targetY;
            this.image = img;
            setImage(img);
        }

        /**
         * Move towards the target location by the specified distance.
         */
        public void moveTowardsTarget(int distance) {

            if (getX() == targetX && getY() == targetY) {
                setRotation(0); // original rotation
            } else if (Math.abs(getX() - targetX) < distance
            && Math.abs(getY() - targetY) < distance) {
                turnTowards(targetX, targetY);
                move(1);
            } else {
                turnTowards(targetX, targetY);
                move(distance);
            }
        }

        public void setTransparency(int t) {
            image.setTransparency(t);
        }

        public boolean isAtTarget() {
            return targetX == getX() && targetY == getY() && getRotation() == 0;
        }
    }
}
